// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// Sine function for QD
//
// Calling Sequence :
// sin(x)
//
// Arguments :
// x : QD number
// 
// This function is based on Bailey's QD, qd_real sin
// http://crd.lbl.gov/~dhbailey/mpdist/
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// make this file : 2011/09/02
// last updating : 2011/09/02
// *****************************************************************
function s = %qd_sin(x)

    if (x == 0) then 
        s = 0.0;
    end

    _2pi = qd(6.283185307179586232e+00,2.449293598294706414e-16,-5.989539619436679332e-33,2.224908441726730563e-49);  // 2*pi
    _pi2 = qd(1.570796326794896558e+00,6.123233995736766036e-17,-1.497384904859169833e-33,5.562271104316826408e-50);    // pi/2
    _pi1024 = qd(3.067961575771282340e-03,1.195944139792337116e-19,-2.924579892303066080e-36,1.086381075061880158e-52); // pi/1024

    // approximately reduce modulo 2*pi
    z = nint_qd(x / _2pi);
    r = x - (_2pi * z);

    // approximately reduce modulo pi/2 and then modulo pi/1024
    j = floor(r.d / _pi2.d + 0.5);
    t = r - (_pi2 * j);
    k = floor(t.d / _pi1024.d + 0.5);
    t = t - (_pi1024 * k);
    abs_k = abs(k);

    // checking an error
    if (j < -2 | j > 2)  then
        error("%qd_sin: Cannot reduce modulo pi/2.");
        // s = tlist(['qd','d'],%nan,%nan,%nan,%nan);
    end

    if (abs_k > 256) then
        error("%qd_sin: Cannot reduce modulo pi/1024.");
        // s = tlist(['qd','d'],%nan,%nan,%nan,%nan);
    end

    if (k == 0) then
        select j
        case 0
            s = sin_taylor_qd(t);
            return;
        case 1
            s = cos_taylor_qd(t);
            return;
        case -1
            s =  -cos_taylor_qd(t);
            return;
        else
            s = -sin_taylor_qd(t);
            return;
        end
    end

    u = cos_table_qd(abs_k);
    v = sin_table_qd(abs_k);

    [sin_t, cos_t] = sincos_taylor_qd(t);
    if (j == 0) then
        if (k > 0) then
            s = u * sin_t + v * cos_t;
        else
            s = u * sin_t - v * cos_t;
        end
    elseif (j == 1) then
        if (k > 0) then
            s = u * cos_t - v * sin_t;
        else
            s = u * cos_t + v * sin_t;
        end
    elseif (j == -1) then
        if (k > 0) then
            s = v * sin_t - u * cos_t;
        else
            s = -u * cos_t - v * sin_t;
        end
    else
        if (k > 0) then
            s = -u * sin_t - v * cos_t;
        else
            s = v * cos_t - u * sin_t;
        end
    end
endfunction
