// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// Tangent function for QD
//
// Calling Sequence :
// tan(x)
//
// Arguments :
// x : QD number
// 
// This function is based on Bailey's QD, qd_real tan
// http://crd.lbl.gov/~dhbailey/mpdist/
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// make this file : 2011/09/03
// last updating : 2011/09/03
// *****************************************************************
function t = %qd_tan(x)
    [sin_x, cos_x] = sincos(x);
    t = sin_x / cos_x;
endfunction

function [sin_x,cos_x] = sincos(x)

    if (x == 0) then
        sin_x = 0.0;
        cos_x = 1.0;
        return;
    end

    _2pi = qd(6.283185307179586232e+00,2.449293598294706414e-16,-5.989539619436679332e-33,2.224908441726730563e-49);  // 2*pi
    _pi2 = qd(1.570796326794896558e+00,6.123233995736766036e-17,-1.497384904859169833e-33,5.562271104316826408e-50);    // pi/2
    _pi1024 = qd(3.067961575771282340e-03,1.195944139792337116e-19,-2.924579892303066080e-36,1.086381075061880158e-52); // pi/1024

    // approximately reduce modulo 2*pi
    z = nint_qd(x / _2pi);
    r = x - (_2pi * z);

    // approximately reduce modulo pi/2 and then modulo pi/1024
    j = floor(r.d / _pi2.d + 0.5);
    t = r - (_pi2 * j);
    k = floor(t.d / _pi1024.d + 0.5);
    t = t - (_pi1024 * k);
    abs_k = abs(k);

    // checking an error
    if (j < -2 | j > 2)  then
        error("sincos: Cannot reduce modulo pi/2.");
        // s = tlist(['qd','d'],%nan,%nan,%nan,%nan);
    end

    if (abs_k > 256) then
        error("sincos: Cannot reduce modulo pi/1024.");
        // s = tlist(['qd','d'],%nan,%nan,%nan,%nan);
    end

    [sin_t, cos_t] = sincos_taylor_qd(t);

    if (k == 0) then
        if (j == 0) then
            sin_x = sin_t;
            cos_x = cos_t;
            return;
        elseif (j == 1)
            sin_x = cos_t;
            cos_x = -sin_t;
            return;
        elseif (j == -1) 
            sin_x = -cos_t;
            cos_x = sin_t;
            return;
        else 
            sin_x = -sin_t;
            cos_x = -cos_t;
            return;
        end
    end

    u = cos_table_qd(abs_k);
    v = sin_table_qd(abs_k);

    if (j == 0) then
        if (k > 0) then
            sin_x = u * sin_t + v * cos_t;
            cos_x = u * cos_t - v * sin_t;
        else 
            sin_x = u * sin_t - v * cos_t;
            cos_x = u * cos_t + v * sin_t;
        end
    elseif (j == 1)
        if (k > 0) then
            cos_x = - u * sin_t - v * cos_t;
            sin_x = u * cos_t - v * sin_t;
        else
            cos_x = v * cos_t - u * sin_t;
            sin_x = u * cos_t + v * sin_t;
        end
    elseif (j == -1)
        if (k > 0) then
            cos_x = u * sin_t + v * cos_t;
            sin_x =  v * sin_t - u * cos_t;
        else
            cos_x = u * sin_t - v * cos_t;
            sin_x = - u * cos_t - v * sin_t;
        end
    else
        if (k > 0) then
            sin_x = - u * sin_t - v * cos_t;
            cos_x = v * sin_t - u * cos_t;
        else
            sin_x = v * cos_t - u * sin_t;
            cos_x = - u * cos_t - v * sin_t;
        end
    end
endfunction

